@@ -10,6 +10,7 @@ type expectation =
   | Failure
   | SuccessOutput of Str.regexp
   | FailureOutput of Str.regexp
+  | SuccessOutputFile of { sample : string; adjust : string -> string; reference : string }
 
 type trace = Off | On of string list
 
@@ -21,29 +22,32 @@ type t = {
   source_teyjus : fname option;
   deps_teyjus : fname list;
   source_dune : fname option;
+  source_json : fname option;
   after: string list;
   typecheck : bool;
   input: fname option;
   expectation : expectation;
   outside_llam : bool;
   trace : string list;
+  legacy_parser : bool;
 }
 
 let tests = ref []
 
 let declare
-    name ~description ?source_elpi ?source_teyjus ?(deps_teyjus=[]) ?source_dune
+    name ~description ?source_elpi ?source_teyjus ?(deps_teyjus=[]) ?source_dune ?source_json
     ?after
     ?(typecheck=true) ?input ?(expectation=Success)
     ?(outside_llam=false)
     ?(trace=Off)
+    ?(legacy_parser=false)
     ~category
     ()
 =
   if List.exists (fun { name = x; _ } -> x = name) !tests then
     failwith ("a test named " ^ name ^ " already exists");
-  begin match source_elpi, source_teyjus, source_dune with
-    | None, None, None -> failwith ("test "^name^" has no sources");
+  begin match source_elpi, source_teyjus, source_dune, source_json with
+    | None, None, None, None-> failwith ("test "^name^" has no sources");
     | _ -> ()
   end;
   tests := { 
@@ -53,21 +57,24 @@ let declare
     source_teyjus;
     deps_teyjus;
     source_dune;
+    source_json;
     after = (match after with None -> [] | Some x -> [x]);
     typecheck;
     input;
     expectation;
     category;
     outside_llam;
+    legacy_parser;
     trace = (match trace with Off -> [] | On l -> "-trace-on" :: l)
   } :: !tests
 
 module SM = Map.Make(String)
 module SS = Set.Make(String)
 
-let get filter =
+let get ~catskip filter =
+  let filtercat x = List.exists ((=) x) catskip in
   let alltests = List.fold_left (fun acc ({ name; _ } as t) -> SM.add name t acc ) SM.empty !tests in
-  let tests = List.filter (fun { name; _ } -> filter ~name) !tests in
+  let tests = List.filter (fun { name; category; _ } -> not (filtercat category) && filter ~name) !tests in
   let testset = List.fold_left (fun acc { name; _ } -> SS.add name acc ) SS.empty tests in
   let deps = List.fold_left (fun acc { after; _ } -> List.fold_right SS.add after acc ) SS.empty tests in
   let to_add = SS.fold (fun n acc -> if SS.mem n testset then acc else SS.add n acc) deps SS.empty in
@@ -158,7 +165,14 @@ let open_log ~executable { Test.name; _ } =
   let name = logdir^"/"^Filename.basename executable^"+"^name^".log" in
   open_file_w name, name
 
-let write (fd,_) s = ignore(Unix.write_substring fd s 0 (String.length s))
+let close_log (fd,_) = Unix.close fd
+
+let open_dummy_log () =
+  let name = Filename.temp_file "elpi" "txt" in
+  open_file_w name, name
+
+let write (fd,f) s =
+  ignore(Unix.write_substring fd s 0 (String.length s))
 
 let open_input name = Unix.openfile name [Unix.O_RDONLY] 0
 
@@ -208,20 +222,20 @@ let wait pid timeout timelog =
     !rc
 
 
-let fork_wait_win ~close_output exe argv env input output timefile timeout =
+let fork_wait_win ~close_output exe argv env input output error timefile timeout =
   ignore close_output;
   let pid =
     Unix.create_process_env
-      exe (Array.of_list argv) env input output output in
+      exe (Array.of_list argv) env input output error in
   wait pid timeout timefile
 
-let fork_wait_unix ~close_output exe argv env input output timefile timeout =
+let fork_wait_unix ~close_output exe argv env input output error timefile timeout =
   let pid = Unix.fork () in
   if pid = 0 then begin
     let _ = Unix.setsid () in
     let runner =
       Unix.create_process_env
-        exe (Array.of_list argv) env input output output in
+        exe (Array.of_list argv) env input output error in
     let _, rc = Unix.waitpid [] runner in
     match rc with
     | Unix.WEXITED n -> exit n
@@ -236,7 +250,7 @@ let fork_wait_unix ~close_output exe argv env input output timefile timeout =
 
 let null = if Sys.win32 then "NUL:" else "/dev/null"
 
-let exec ~timeout ?timetool ~env ~log:(output,oname) ?(input=null) ~executable ~args ?(close_output=true) () =
+let exec ~timeout ?timetool ~env ~log:(output,oname) ?(input=null) ?(error=output,oname) ~executable ~args ?(close_output=true) () =
   Sys.catch_break true;
   let exe, argv, timefile =
     match timetool with
@@ -247,8 +261,8 @@ let exec ~timeout ?timetool ~env ~log:(output,oname) ?(input=null) ~executable ~
                    executable :: args), timefile in
   Unix.handle_unix_error (fun () ->
     let input = open_input input in
-    if Sys.win32 then fork_wait_win ~close_output exe argv env input output timefile timeout
-    else fork_wait_unix  ~close_output exe argv env input output timefile timeout)
+    if Sys.win32 then fork_wait_win ~close_output exe argv env input output (fst error) timefile timeout
+    else fork_wait_unix  ~close_output exe argv env input output (fst error) timefile timeout)
     ()
 
 let with_log (_,log) f =
@@ -261,14 +275,49 @@ let with_log (_,log) f =
 
 let option_map f = function None -> None | Some x -> Some (f x)
 
+let strip_cwd file =
+  let name = Filename.temp_file "elpi" "txt" in
+  let oc = open_out name in
+  let ic = open_in file in
+  let rex = Str.regexp_string (Str.global_replace (Str.regexp_string "\\") "/" (Sys.getcwd () ^ Filename.dir_sep)) in
+  try
+    while true do
+      let l = input_line ic in
+      let l = Str.global_replace rex "" l in
+      output_string oc (l ^ "\n")
+    done;
+    name
+  with End_of_file ->
+    close_in ic;
+    close_out oc;
+    name
+    
 end
 
+let match_file ~log file adjust reference =
+  let file = adjust file in
+  Util.write log (Printf.sprintf "Diffing %s against %s\n" file reference);
+  match
+    Util.exec ~timeout:5.0 ~env:(Unix.environment ())
+      ~log ~close_output:false
+      ~executable:"diff" ~args:["-u";"--strip-trailing-cr";reference;file] ()
+  with
+  | Util.Exit(0,_,_) -> true
+  | Util.Exit(n,_,_) ->
+    Util.write log (Printf.sprintf "Exit code: %d\n" n);
+    Util.write log (Printf.sprintf "Promotion: cp %s %s\n" file reference);
+    false
+  | _ -> false
+
 
 module Elpi = struct
 
 let is_elpi =
-  let rex = Str.regexp "elpi.*" in
-  fun s -> Str.string_match rex (Filename.basename s) 0
+  let rex1 = Str.regexp "elpi.*$" in
+  let rex2 = Str.regexp "elpi-trace-elaborator.*$" in
+  fun s ->
+    let s = Filename.basename s in
+    Str.string_match rex1 s 0 && not (Str.string_match rex2 s 0)
 
 let read_time input_line =
   let time = ref 0.0 in
@@ -281,13 +330,18 @@ let read_time input_line =
   with End_of_file -> !time
 
 let match_rex rex input_line =
-  let b = ref false in
-  try while true do
+  let b = Buffer.create 100 in
+  begin try
+    while true do
     let l = input_line () in
-    try ignore(Str.search_forward rex l 0); b := true
-    with Not_found -> ()
-  done; !b
-  with End_of_file -> !b
+    Buffer.add_string b l;
+    Buffer.add_string b "\n";
+    done
+  with End_of_file -> () end;
+  let s = Buffer.contents b in
+  let s = Str.global_replace (Str.regexp_string "\r") "" s in
+  try ignore(Str.search_forward rex s 0); true
+  with Not_found -> false
 
 let read_tctime input_line =
   let time = ref 0.0 in
@@ -299,9 +353,20 @@ let read_tctime input_line =
   done; !time
   with End_of_file -> !time
 
+let legacy_parser_available executable =
+  let log = Util.open_dummy_log () in
+  let env = Unix.environment () in
+  match
+    Util.exec ~executable ~timeout:1.0 ~env ~log ~args:["-legacy-parser-available"] ()
+  with
+  | Util.Exit(0,_,_) -> true
+  | _ -> false
+
 let () = Runner.declare
-  ~applicable:begin fun ~executable { Test.source_elpi; _ } ->
-    if is_elpi executable && source_elpi <> None then Runner.Can_run_it
+  ~applicable:begin fun ~executable { Test.source_elpi; legacy_parser; _ } ->
+    if is_elpi executable && source_elpi <> None && 
+      (not legacy_parser || legacy_parser_available executable)
+      then Runner.Can_run_it
     else Runner.Not_for_me
   end
   ~run:begin fun ~executable ~timetool ~timeout ~env ~sources test ->
@@ -314,53 +379,107 @@ let () = Runner.declare
     Util.write log (Printf.sprintf "executable: %s\n" executable);
     let executable_stuff = Filename.dirname executable ^ "/../lib/elpi/" in
 
-    let { Test.expectation; input; outside_llam ; typecheck; trace; _ } = test in
+    let { Test.expectation; input; outside_llam ; typecheck; trace; legacy_parser; _ } = test in
     let input = Util.option_map (fun x -> sources^x) input in
     let args = ["-test";"-I";executable_stuff;"-I";sources;source] @ trace in
     let args =
       if typecheck then args
       else "-no-tc" :: args in
+    let args =
+      if not legacy_parser then args
+      else "-legacy-parser" :: args in
     let args =
       if outside_llam then "-delay-problems-outside-pattern-fragment"::args
       else args in
     Util.write log (Printf.sprintf "args: %s\n" (String.concat " " args));
 
     let rc =
-      match expectation, Util.exec ~timeout ~timetool ?input ~executable ~env ~log ~args () with
-      | Test.Success, Util.Exit(0,walltime,mem) ->
-        Runner.Success { walltime; typechecking = Util.with_log log read_tctime; execution = Util.with_log log read_time; mem }
-      | Test.Failure, Util.Exit(0,walltime,mem) ->
-        Runner.Failure { walltime; typechecking = Util.with_log log read_tctime; execution = Util.with_log log read_time; mem }
-      | Test.Success, Util.Exit(_,walltime,mem) ->
-        Runner.Failure { walltime; typechecking = Util.with_log log read_tctime; execution = walltime; mem }
-      | Test.Failure, Util.Exit(_,walltime,mem) ->
-        Runner.Success { walltime; typechecking = Util.with_log log read_tctime; execution = Util.with_log log read_time; mem }
-      | Test.Success, Util.Timeout ->
-        Runner.Timeout timeout
-      | Test.Failure, Util.Timeout ->
-        Runner.Timeout timeout
-      | Test.FailureOutput _, Util.Exit(0,walltime,mem) ->
-           Runner.Failure { walltime; typechecking = Util.with_log log read_tctime; execution = walltime; mem }
-     | Test.SuccessOutput rex, Util.Exit(0,walltime,mem) ->
-          if Util.with_log log (match_rex rex) then
-            Runner.Success { walltime; typechecking = Util.with_log log read_tctime; execution = Util.with_log log read_time; mem }
-          else
-            Runner.Failure { walltime; typechecking = Util.with_log log read_tctime; execution = walltime; mem }
-      | Test.SuccessOutput _, Util.Exit(_,walltime,mem) ->
-          Runner.Failure { walltime; typechecking = Util.with_log log read_tctime; execution = walltime; mem }
-      | Test.FailureOutput rex, Util.Exit(_,walltime,mem) ->
-          if Util.with_log log (match_rex rex) then
-            Runner.Success { walltime; typechecking = Util.with_log log read_tctime; execution = Util.with_log log read_time; mem }
-          else
-            Runner.Failure { walltime; typechecking = Util.with_log log read_tctime; execution = walltime; mem }
-      | (Test.FailureOutput _ | Test.SuccessOutput _), Util.Timeout ->
-        Runner.Timeout timeout
+      let outcome = Util.exec ~timeout ~timetool ~close_output:false ?input ~executable ~env ~log ~args () in
+      let typechecking = Util.with_log log read_tctime in
+      let execution = Util.with_log log read_time in
+
+      match outcome with
+      | Util.Exit(0,walltime,mem) ->
+        begin match expectation with
+        | Test.Success -> Runner.Success { walltime; typechecking; execution; mem }
+        | Test.SuccessOutput rex when Util.with_log log (match_rex rex) -> Runner.Success { walltime; typechecking; execution; mem }
+        | Test.SuccessOutputFile { sample; adjust; reference } when match_file ~log sample adjust (sources^"/"^reference) -> Runner.Success { walltime; typechecking; execution; mem }
+        | _ -> Runner.Failure { walltime; typechecking; execution; mem }
+        end
+
+      | Util.Exit(_,walltime,mem) ->
+        begin match expectation with
+        | Test.Failure -> Runner.Success { walltime; typechecking; execution; mem }
+        | Test.FailureOutput rex when Util.with_log log (match_rex rex) -> Runner.Success { walltime; typechecking; execution; mem }
+        | _ -> Runner.Failure { walltime; typechecking; execution; mem }
+        end
+
+      | Util.Timeout -> Runner.Timeout timeout
+        
     in
+    Util.close_log log;
     Runner.(Done { Runner.rc; executable; test; log = snd log })
   end
 
 end
 
+let match_rex rex input_line =
+  let b = ref false in
+  try while true do
+    let l = input_line () in
+    try ignore(Str.search_forward rex l 0); b := true
+    with Not_found -> ()
+  done; !b
+  with End_of_file -> !b
+
+module ElpiTraceElab = struct
+  let is_elpi_elab =
+    let rex = Str.regexp "elpi-trace-elaborator" in
+    fun s -> Str.string_match rex (Filename.basename s) 0
+  
+  let () = Runner.declare
+    ~applicable:(fun ~executable { Test.source_json; } ->
+      if is_elpi_elab executable && source_json <> None then
+        Runner.Can_run_it
+      else Runner.Not_for_me)
+    ~run:(fun ~executable ~timetool ~timeout ~env ~sources test ->
+    if not (Sys.file_exists executable) then Runner.Skipped
+    else
+      let log = Util.open_log ~executable test in
+      Util.write log (Printf.sprintf "executable: %s\n" executable);
+      let source =
+        match test.Test.source_json with Some x -> x | _ -> assert false in
+      let input = sources ^ source in
+      Util.write log (Printf.sprintf "input: %s\n" input);
+      let outcome, outcomey =
+        match test.Test.expectation with
+        | Test.SuccessOutputFile { sample } ->
+            let output_file = Util.open_file_w sample, sample in
+            let output_file_tmp = Util.open_dummy_log () in
+            Util.write log (Printf.sprintf "output: %s\n" (snd output_file));
+            let outcome = Util.exec ~timeout ~timetool ~input ~executable ~env ~error:log ~log:output_file_tmp ~args:[] () in
+            let outcomey = Some (Util.exec ~timeout ~timetool ~input:(snd output_file_tmp) ~executable:"ydump" ~env ~log:output_file ~args:[] ()) in
+            outcome, outcomey
+        | Test.FailureOutput _ ->
+            let outcome = Util.exec ~timeout ~timetool ~input ~executable ~env ~log ~close_output:false ~args:[] () in
+            outcome, None
+        | _ -> Printf.eprintf "ElpiTraceElab only supoprts SuccessOutputFile / FailureOutput tests"; exit 1 in
+      let typechecking = 0.0 in
+      let execution = 0.0 in
+      let rc =
+        match outcome, outcomey, test.Test.expectation with
+        | Util.Exit(0,walltime,mem), Some(Util.Exit(0,_,_)), Test.SuccessOutputFile { sample; adjust; reference } when match_file ~log sample adjust (sources^"/"^reference) ->
+            Runner.Success { walltime; typechecking; execution; mem }
+        | Util.Exit(m,walltime,mem), None, Test.FailureOutput rex when m != 0 && Util.with_log log (match_rex rex) ->
+            Runner.Success { walltime; typechecking; execution; mem }
+        | _ -> Runner.Failure { walltime = 0.0; typechecking; execution = 0.0; mem = 0 }
+      in
+      Util.close_log log;
+      Runner.(Done { Runner.rc; executable; test; log = snd log })
+      )
+
+end
+
 module Teyjus = struct
 
 let assert_ok = function
@@ -412,25 +531,25 @@ let is_tjsim =
     let args = ["-m";"1";"-b";"-s";"main.";"-p";dir^"/";modname] in
 
     Util.write log (Printf.sprintf "%s %s\n" executable (String.concat " " args));
-    let rc = Util.exec ~timeout ~timetool ~executable ?input ~env ~log ~args ~close_output:false () in
+    let outcome = Util.exec ~timeout ~timetool ~executable ?input ~env ~log ~args ~close_output:false () in
+    let typechecking = 0.0 in
     let rc =
-      match expectation, rc with
-      | Test.Success, Util.Exit(0,walltime,mem) ->
-        Runner.Success { walltime; typechecking = 0.0; execution = walltime; mem }
-      | Test.Failure, Util.Exit(0,walltime,mem) ->
-        Runner.Failure { walltime; typechecking = 0.0; execution = walltime; mem }
-      | Test.Success, Util.Exit(_,walltime,mem) ->
-        Runner.Failure { walltime; typechecking = 0.0; execution = walltime; mem }
-      | Test.Failure, Util.Exit(_,walltime,mem) ->
-        Runner.Success { walltime; typechecking = 0.0; execution = walltime; mem }
-      | Test.Success, Util.Timeout ->
-        Runner.Timeout timeout
-      | Test.Failure, Util.Timeout ->
-        Runner.Timeout timeout
-      | (Test.SuccessOutput _ | Test.FailureOutput _), Util.Exit(_,walltime,mem) ->
-        Runner.Success { walltime; typechecking = 0.0; execution = walltime; mem }
-      | (Test.SuccessOutput _ | Test.FailureOutput _), Util.Timeout ->
-        Runner.Timeout timeout
+
+      match outcome with
+      | Util.Exit(0,walltime,mem) -> let execution = walltime in
+        begin match expectation with
+        | Test.Success -> Runner.Success { walltime; typechecking; execution; mem }
+        | _ -> Runner.Failure { walltime; typechecking; execution; mem }
+        end
+
+      | Util.Exit(_,walltime,mem) -> let execution = walltime in
+        begin match expectation with
+        | Test.Failure -> Runner.Success { walltime; typechecking; execution; mem }
+        | _ -> Runner.Failure { walltime; typechecking; execution; mem }
+        end
+
+      | Util.Timeout -> Runner.Timeout timeout
+
     in
     Runner.Done { Runner.rc; executable; test; log = snd log }
   end
@@ -447,15 +566,6 @@ let is_dune_src = function
   | None -> false
   | Some s -> Filename.check_suffix s ".ml"
 
-let match_rex rex input_line =
-  let b = ref false in
-  try while true do
-    let l = input_line () in
-    try ignore(Str.search_forward rex l 0); b := true
-    with Not_found -> ()
-  done; !b
-  with End_of_file -> !b
-
 let () = Runner.declare
   ~applicable:begin fun ~executable { source_dune; _ } ->
     if is_dune executable && is_dune_src source_dune then Runner.Can_run_it else Runner.Not_for_me
@@ -474,31 +584,25 @@ let () = Runner.declare
     let args = ["exec"; sources ^ "/" ^ source; "--"; "-I"; "src/"] in
     Util.write log (Printf.sprintf "args: %s\n" (String.concat " " args));
     let rc =
-      match expectation, Util.exec ~timeout ~timetool ?input ~executable ~env ~log ~args () with
-      | Test.Success, Util.Exit(0,walltime,mem) ->
-        Runner.Success { walltime; typechecking = 0.0; execution = 0.0; mem }
-      | Test.Success, Util.Exit(_,walltime,mem)->
-        Runner.Failure { walltime; typechecking = 0.0; execution = 0.0; mem }
-      | Test.Failure, Util.Exit(0,walltime,mem) ->
-        Runner.Failure { walltime; typechecking = 0.0; execution = 0.0; mem }
-      | Test.Failure, Util.Exit(_,walltime,mem)->
-        Runner.Success { walltime; typechecking = 0.0; execution = 0.0; mem }
-      | Test.SuccessOutput rex, Util.Exit(0,walltime,mem) ->
-          if Util.with_log log (match_rex rex) then
-            Runner.Success { walltime; typechecking = 0.0; execution = 0.0; mem }
-          else
-            Runner.Failure { walltime; typechecking = 0.0; execution = walltime; mem }
-      | Test.FailureOutput _, Util.Exit(0,walltime,mem) ->
-            Runner.Failure { walltime; typechecking = 0.0; execution = walltime; mem }
-      | Test.SuccessOutput _, Util.Exit(_,walltime,mem) ->
-            Runner.Failure { walltime; typechecking = 0.0; execution = walltime; mem }
-      | Test.FailureOutput rex, Util.Exit(_,walltime,mem) ->
-          if Util.with_log log (match_rex rex) then
-            Runner.Success { walltime; typechecking = 0.0; execution = 0.0; mem }
-          else
-            Runner.Failure { walltime; typechecking = 0.0; execution = walltime; mem }
-      | _, Util.Timeout ->
-        Runner.Timeout timeout
+      let outcome = Util.exec ~timeout ~timetool ?input ~executable ~env ~log ~args () in
+      let typechecking = 0.0 in
+      match outcome with
+      | Util.Exit(0,walltime,mem) -> let execution = walltime in
+        begin match expectation with
+        | Test.Success -> Runner.Success { walltime; typechecking; execution; mem }
+        | Test.SuccessOutput rex when Util.with_log log (match_rex rex) -> Runner.Success { walltime; typechecking; execution; mem }
+        | _ -> Runner.Failure { walltime; typechecking; execution; mem }
+        end
+
+      | Util.Exit(_,walltime,mem) -> let execution = walltime in
+        begin match expectation with
+        | Test.Failure -> Runner.Success { walltime; typechecking; execution; mem }
+        | Test.FailureOutput rex when Util.with_log log (match_rex rex) -> Runner.Success { walltime; typechecking; execution; mem }
+        | _ -> Runner.Failure { walltime; typechecking; execution; mem }
+        end
+
+      | Util.Timeout -> Runner.Timeout timeout
+
     in
     Runner.(Done { Runner.rc; executable; test; log = snd log })
 
